;;; -*- mode:scheme; coding:utf-8; -*-
;;;
;;; sagittarius/bv-string.scm - String as bytevector
;;;  
;;;   Copyright (c) 2010-2015  Takashi Kato  <ktakashi@ymail.com>
;;;   
;;;   Redistribution and use in source and binary forms, with or without
;;;   modification, are permitted provided that the following conditions
;;;   are met:
;;;   
;;;   1. Redistributions of source code must retain the above copyright
;;;      notice, this list of conditions and the following disclaimer.
;;;  
;;;   2. Redistributions in binary form must reproduce the above copyright
;;;      notice, this list of conditions and the following disclaimer in the
;;;      documentation and/or other materials provided with the distribution.
;;;  
;;;   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;;   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;;   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;;   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;;   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;;   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
;;;   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
;;;   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
;;;   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
;;;   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;;   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;;  

;; reader macro #*"string"
;; this is read as bytevector which content is the same as 
;; (string->utf8 "string")
;; TODO
;;  we may want to support other codings. e.g.) #*(utf16 "string")
(library (sagittarius bv-string)
    (export :export-reader-macro)
    (import (rnrs)
	    (core errors) ;; for lexical-violation
	    (sagittarius reader))

  (define-dispatch-macro #\# #\*
    (bv-string-reader iport subchar param)
    ;; (define string-reader (get-macro-character #\"))
    (case (lookahead-char iport)
      ((#\")
       ;; let built-in procedure handle it
       ;; TODO this doesn't secure if '"' reader macro is overwritten
       ;;      we may need read-string or something
       (string->utf8 (read iport)))
      (else
       (lexical-violation 'bv-string-reader "#* expects #\""))))


)
